/***
* munpos.c
*
* Copyright (c) 2005 - 2015, Wolfgang Astl. All rights reserved.
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
* MA 02110-1301 USA
*
*/

/* Created   2005/11/13 */
/* Last Mod. 2015/12/14 */


#include <stddef.h>
#include <math.h>
#include "munpos.h"
#include <stdio.h>

#define PI     3.14159265358979323846264338327950
#define ARC30  (PI / 6.)
#define ARC90  (PI / 2.)
#define ARC270 (PI / 2. * 3.)
#define ARC360 (PI * 2.)
#define RAD    (PI / 180.)
#define TINY 1e-9
#define TINY_RAD (TINY * RAD)
#define TRUE 1
#define FALSE !TRUE

#define PF(x) (double) ((x) < PI ? (ARC90 - (x)) / ARC90 : ((x) - ARC270) / ARC90)
#define PD_BUG 1

/*============================================================
// private:
//==========================================================*/

/* misc
//----------------------------------------------------------*/

static double norm2PI(double x)
{
  x = fmod(x, ARC360);
  if(x < 0)
    x += ARC360;
  if(x >= ARC360 - TINY_RAD)
    x = 0.0;
  return(x);
}

static double normPI(double x)
{
  x = fmod(x, ARC360);
  if(x < 0)
    x += ARC360;
  if(x >= PI)
    x -= ARC360;
  return(x);
}

static double normARC90(double x)
{
  x = normPI(x);
  if(x > ARC90)
    x = PI - x;
  if(x < -ARC90)
    x = -PI - x;
  if(fabs(fabs(x) - ARC90) < TINY_RAD) 
  {
    if(x < 0.0)
      x = -ARC90 + TINY_RAD;
    else
      x = +ARC90 - TINY_RAD;
  }
  return(x);
}

static double Trunc(double x)
{
  if(x < 0.0)
    x = ceil(x);
  else
    x = floor(x);
  return(x);
}

static double ecl2ra(double eps, double lon, double lat)
{
  double ra = atan2(sin(lon) * cos(eps) - tan(lat) * sin(eps), cos(lon));
  if(ra < 0.)
    ra += ARC360;
  return(ra);
}

static double ecl2de(double eps, double lon, double lat)
{
  return asin(sin(lon) * cos(lat) * sin(eps) + sin(lat) * cos(eps));
}

static void cotrans(double eps, double lon, double lat, double *gc)
{
  gc[0] = ecl2ra(eps, lon, lat);  
  gc[1] = ecl2de(eps, lon, lat);
}

static double equ2az(double lst, double phi, double ra, double de)
{
  double ha = lst - ra;
  double az = atan2(sin(ha), cos(ha) * sin(phi) - tan(de) * cos(phi)) + PI;
  if(az < 0.)
    az += ARC360;
  return az;
}

static double equ2alt(double lst, double phi, double ra, double de)
{
  double ha = lst - ra;
  return asin(sin(phi) * sin(de) + cos(phi) * cos(de)  * cos(ha));
}

static double quart(double mp, double q0, double q1)
{
  int q;
  double qq[4], tmp;
  mp = norm2PI(mp);
  qq[0] = q0;
  qq[1] = q1;
  qq[2] = q0 + PI;
  qq[3] = q1 + PI;
  q = (int)(mp / ARC90 + TINY_RAD);       /* q = quarter 0 .. 3 */
  tmp = norm2PI(qq[(q + 1) & 3] - qq[q]); /* width */
  return qq[q] + ((mp - q * ARC90) / ARC90) * tmp;
}

static int pdist(MUNPOS *m, double ra, double de, double *vect)
{
  vect[0] = normPI(ra - m->lst);         /* umd */
  vect[1] = -normPI(PI - vect[0]);       /* lmd */
  vect[4] = tan(m->phi) * tan(de);       /* sin AD */
  if(fabs(vect[4]) > 1.0)
    return TRUE;
  else 
  {
    vect[4] = asin(vect[4]);             /* ad */
    vect[2] = ARC90 + vect[4];           /* dsa */
    vect[3] = ARC90 - vect[4];           /* nsa */
    if(fabs(vect[0]) > vect[2])
      vect[5] = vect[1] / vect[3];       /* proportionaldistance */
    else
      vect[5] = vect[0] / vect[2];
  }
  return FALSE;
}

double gc2lon(double eps, double OA, double psi)
{
  double x, y, lon;
  x = cos(OA) * cos(eps);
  if(psi != 0.)
    x -= tan(psi) * sin(eps);
  y = sin(OA);
  if(fabs(x) < TINY_RAD)
    x = 0.;
  if(fabs(y) < TINY_RAD)
    y = 0.;
  if(x == 0. && y == 0.)
    lon = floor((OA + TINY_RAD) / ARC90) * ARC90;
  else
    lon = atan2(y, x);
  if(lon < 0)
    lon += ARC360;
  return(lon);
}


/* calc mundane positions from rectascension and declination
//----------------------------------------------------------*/

/* MC - equal */
static double mp_mceq(MUNPOS *m, double ra, double de)
{
  return ecl2ra(-m->eps, ra, de) - ARC90 - m->mc;
}

/* AC - equal */
static double mp_aceq(MUNPOS *m, double ra, double de)
{
  return ecl2ra(-m->eps, ra, de) - m->ac;
}

/* Porphyry */
static double mp_porp(MUNPOS *m, double ra, double de)
{
  double mp, lon, nsa;
  lon = norm2PI(ecl2ra(-m->eps, ra, de) - m->ac);
  nsa = norm2PI(PI + m->mc - m->ac);
  mp = 0;
  if(lon >= PI) 
  {
    lon -= PI;
    mp += PI;
  }
  mp += lon < nsa ? ARC90 / nsa * lon : ARC90 / (PI - nsa) * (lon - nsa) + ARC90;
  return mp;
}

/* Meridian */
static double mp_meri(MUNPOS *m, double ra, double de)
{
  return ra - m->lst - ARC90;
}

/* Poli-equatorial */
static double mp_poli(MUNPOS *m, double ra, double de)
{
  return ra - m->raac;
}

/* Alcabitius */
static double mp_alca(MUNPOS *m, double ra, double de)
{
  double mp, ix, nsa;
  ix = norm2PI(ra - m->raac);        
                                     
  nsa = norm2PI(m->lst + PI - m->raac);   
  mp = 0.0;
  if(ix >= PI)
  {
    ix -= PI;   
    mp += PI; 
  }
  if(ix < nsa)
    mp += ARC90 / nsa * ix;
  else
    mp += ARC90 / (PI - nsa) * (ix - nsa) + ARC90;
  return mp;
}

/* Campanus */
static double mp_camp(MUNPOS *m, double ra, double de)
{
  return ecl2ra(m->phi, ra - ARC90 - m->lst, de);
}

/* Regiomontanus */
static double mp_regi(MUNPOS *m, double ra, double de)
{
  double mp = mp_camp(m, ra, de);
  return atan2(cos(mp) * cos(m->phi), -sin(mp)) - ARC90;
}

/* Abenragel (APC) */
static double mp_haly(MUNPOS *m, double ra, double de)
{
  double mp, psi, ad, na, sin_ad;
  mp = mp_regi(m, ra, de);            /* regiomontanus mundane position */
  psi = atan(cos(mp) * tan(m->phi));  /* regiomontanus pole */
  if(fabs(sin_ad = tan(m->de) * tan(psi)) > 1.0)
    return mpERROR;
  ad = asin(sin_ad);
  ra = m->lst + ARC90 + mp + ad;
  na = PI - 2 * m->ad;
  ra = norm2PI(ra - m->lst - PI + na / 2.0);
  mp = ra < na ? PI / na * ra : PI / (ARC360 - na) * (ra - na) + PI;
  return mp;
}

/* Placidus */
static double mp_plac(MUNPOS *m, double ra, double de)
{
  double pd[6];
  if(pdist(m, ra, de, pd))
    return mpERROR;
  if(fabs(pd[0]) > pd[2])
    return  ARC90 + ARC90 * pd[5];
  else
    return ARC270 + ARC90 * pd[5];
}

/* Morinus */
static double mp_mori(MUNPOS *m, double ra, double de)
{
  double lon = ecl2ra(-m->eps, ra, de);
  return -atan2(cos(m->eps) * cos(lon), sin(lon)) - m->lst;
}

/* Topocentric */
static double mp_topo(MUNPOS *m, double ra, double de)
{
  double mp, pd[6], sin_ad, ad, psi, ra0;
  int i;
  if(pdist(m, ra, de, pd))
    return mpERROR;
  psi = atan(tan(m->phi) * fabs(pd[5]));  /* topocentric pole */
  ad = asin(tan(psi) * tan(de));          /* topocentric ad */
  if(pd[0] > 0.0)  /* east */
    mp = ra - ad - ARC90 - m->lst;  
  else             /* west */
    mp = ra + ad - ARC90 - m->lst;     

  sin_ad = tan(m->phi) * tan(de);
  for(i = 0; i < 50; i++) 
  {
    mp = norm2PI(mp);
    ra0 = norm2PI(m->lst + ARC90 + mp + asin(PF(mp) * sin_ad));
    if(fabs(ra0 - ra) < 0.0000002 * RAD)
      break;
    mp += ra - ra0;
  }
  return mp;
}

/* Koch */
static double mp_koch(MUNPOS *m, double ra, double de)
{
  double sin_ad, ad, umd, lmd;
  if(m->flags & mpERR_DCP) 
    return mpERROR;
  sin_ad = tan(m->phi) * tan(de);
  if(fabs(sin_ad) > 1.0)
    return mpERROR;
  ad = asin(sin_ad);
  umd = normPI(ra - m->lst);
  lmd = normPI(umd + PI);
  if(umd >= 0.0) /* east */
    return ((umd - ad + m->ad) / (m->ad + ARC90) - 1) * ARC90;
  else           /* west */
    return ((lmd + ad + m->ad) / (m->ad + ARC90) + 1) * ARC90;
}


/* calc OA and pole from mundane position mp
//----------------------------------------------------------*/

/* MC - equal */
static void gc_mceq(MUNPOS *m, double mp, double *gc)
{
  double lon = m->mc + ARC90 + mp;
  gc[0] = ARC90 - atan2(cos(m->eps) * cos(lon), sin(lon)); /* OA */
  gc[1] = atan(-cos(gc[0]) * tan(m->eps));                 /* pole */
}

/* AC - equal */
static void gc_aceq(MUNPOS *m, double mp, double *gc)
{
  double lon = m->ac + mp;
  gc[0] = ARC90 - atan2(cos(m->eps) * cos(lon), sin(lon));
  gc[1] = atan(-cos(gc[0]) * tan(m->eps));        
}

/* Porphyry */
static void gc_porp(MUNPOS *m, double mp, double *gc)
{
  double lon = quart(mp, m->ac, m->mc + PI);
  gc[0] = ARC90 - atan2(cos(m->eps) * cos(lon), sin(lon));
  gc[1] = atan(-cos(gc[0]) * tan(m->eps));            
}

/* Neridian */
static void gc_meri(MUNPOS *m, double mp, double *gc)
{
  gc[0] = m->lst + ARC90 + mp;
  gc[1] = 0.0;
}

/* Poli-Equatorial */
static void gc_poli(MUNPOS *m, double mp, double *gc)
{
  gc[0] = m->raac + mp;
  gc[1] = 0.0;
}

/* Alcabitius */
static void gc_alca(MUNPOS *m, double mp, double *gc)
{
  gc[0] = quart(mp, m->raac, m->lst + PI);
  gc[1] = 0.0;
}

/* Campanus */
static void gc_camp(MUNPOS *m, double mp, double *gc)
{
  gc[0] = atan2(cos(m->phi) * cos(mp), -sin(mp)) + m->lst;
  gc[1] = atan(sin(gc[0] - m->lst) * tan(m->phi));
}

/* Regiomontanus */
static void gc_regi(MUNPOS *m, double mp, double *gc)
{
  gc[0] = m->lst + ARC90 + mp;
  gc[1] = atan(cos(mp) * tan(m->phi));
}

/* Abenragel, APC */
static void gc_haly(MUNPOS *m, double mp, double *gc)
{
  double nsa, ra;
  mp = norm2PI(mp);
  nsa = ARC90 - m->ad;
  if(mp < PI)
    ra = (mp / ARC90 - 1.0) * nsa + PI;
  else
    ra = (mp / ARC90 - 3.0) * (PI - nsa);
  mp = mp_regi(m, m->lst + ra, m->de); /* regiomontanus mundane position */
  gc[0] = m->lst + ARC90 + mp;         /* regiomontanus OA */
  gc[1] = atan(cos(mp) * tan(m->phi)); /* regiomontanus pole */
  return;
}

/* Placidus */
static void gc_plac(MUNPOS *m, double mp, double *gc)
{
  double tan_dec, pd, tmp, cusp, ad;
  int i;
  mp = norm2PI(mp);
  gc[0] = m->lst + ARC90 + mp;  /* OA */
  tan_dec = tan(m->eps);
  pd = PF(mp);
  tmp = (cusp = 0.) + 1.;
  for(i = 0; i < 10; i++) 
  {
    ad = asin(tan_dec * m->tan_phi);       
    gc[1] = atan2(sin(ad * pd), tan_dec);   
    if(fabs(cusp - tmp) < (0.00001 * RAD))  
      break;
    tmp = cusp;
    cusp = gc2lon(m->eps, gc[0], gc[1]);         
    tan_dec = tan(asin(m->sin_eps * sin(cusp))); 
  }
}

/* Morinus */
static void gc_mori(MUNPOS *m, double mp, double *gc)
{
  gc[0] = m->lst + ARC90 + mp;
  gc[1] = atan(-cos(gc[0]) * tan(m->eps));
}

/* Topocentric */
static void gc_topo(MUNPOS *m, double mp, double *gc)
{
  mp = norm2PI(mp);
  gc[0] = mp + ARC90 + m->lst;
  gc[1] = atan(PF(mp) * tan(m->phi));
}

/* Koch */
static void gc_koch(MUNPOS *m, double mp, double *gc)
{
  mp = norm2PI(mp);
  if(mp < ARC90) 
  {
    gc[0] = m->lst + ARC90  + (mp / ARC90)    * (ARC90 + m->ad);
    gc[1] = m->phi;
  }
  else if(mp < ARC270) 
  {
    gc[0] = m->lst + ARC270 + (mp / ARC90 - 2) * (ARC90 + m->ad);
    gc[1] = -m->phi;
  }
  else 
  {
    gc[0] = m->lst + ARC90  + (mp / ARC90 - 4) * (ARC90 + m->ad);
    gc[1] = m->phi;
  }
  gc[0] = norm2PI(gc[0]);
}


static double gc2(MUNPOS *m, double ra, double de, double *gc)
{
  double mp, tan_psi, sin_ad;
  if((mp = m->fp_mp(m, ra, de)) == mpERROR)
    return mpERROR;
  switch(m->hsys)
  {
    case mpPLAC:
    case mpKUEH:
      gc[1] = fabs(de) < TINY_RAD ? PF(mp) * tan(m->phi) : sin(ra - mp - ARC90 - m->lst) / tan(de);
      gc[0] = m->lst + ARC90 + mp;
      break;

    case mpPOPA:
      gc[1] = PF(mp) * tan(m->phi);
      if(fabs(sin_ad = gc[1] * tan(de)) > 1.0)
        return mpERROR;
      gc[0] = ra - asin(sin_ad);
      break;

    default:
      if(m->flags & mpERRORS)
        return mpERROR;
      m->fp_gc(m, mp, gc);
      gc[1] = tan(gc[1]);
  }
  return mp;
}

static double direct(MUNPOS *m, double *pro, double mp, double *gc)
{
  double sin_ad, ip, arc;
  mp = norm2PI(mp);
  gc[0] = norm2PI(gc[0]);
  if(m->hsys == mpPLAC)
  {
    if(fabs(sin_ad = tan(m->phi) * tan(pro[1])) > 1.0)
      return(mpERROR);  
    ip = gc[0] + PF(mp) * asin(sin_ad);
  }
  else
  {
    if(fabs(sin_ad = gc[1] * tan(pro[1])) > 1.0)
      return(mpERROR);  
    ip = gc[0] + asin(sin_ad);
  }
  arc = normPI(pro[0] - ip);
  return arc;
}

static double dir1(MUNPOS *m, double *pro, double mp)
{
  double gc[2];
  if((m->flags & mpERRORS) || m->hsys == mpKUEH || m->hsys == mpPOPA)
    return mpERROR;
  m->fp_gc(m, mp, gc);
  gc[1] = tan(gc[1]);
  return direct (m, pro, mp, gc);
}
    
static double dir2(MUNPOS *m, double *pro, double *sig)
{
  double mp, gc[2];
  if((mp = gc2(m, sig[0], sig[1], gc)) == mpERROR)
    return mpERROR;  
  else
    return direct(m, pro, mp, gc);  
}


/* init functors
//----------------------------------------------------------*/
static int initFP(MUNPOS *m, unsigned hsys)
{
  m->flags = hsys;  /* & ~0x7f */
  m->hsys = hsys & 0x7f;       
  switch(m->hsys) 
  {
    case mpMCEQ:    m->fp_mp = mp_mceq; m->fp_gc = gc_mceq; break;
    case mpACEQ:    m->fp_mp = mp_aceq; m->fp_gc = gc_aceq; break;
    case mpPORP:    m->fp_mp = mp_porp; m->fp_gc = gc_porp; break;
    case mpMERI:    m->fp_mp = mp_meri; m->fp_gc = gc_meri; break;
    case mpPOLI:    m->fp_mp = mp_poli; m->fp_gc = gc_poli; break;
    case mpALCA:    m->fp_mp = mp_alca; m->fp_gc = gc_alca; break;
    case mpCAMP:    m->fp_mp = mp_camp; m->fp_gc = gc_camp; break;
    case mpREGI:    m->fp_mp = mp_regi; m->fp_gc = gc_regi; break;
    case mpHALY:    m->fp_mp = mp_haly; m->fp_gc = gc_haly; break;
    case mpPLAC:    m->fp_mp = mp_plac; m->fp_gc = gc_plac; break;
    case mpTOPO:    m->fp_mp = mp_topo; m->fp_gc = gc_topo; break;
    case mpKOCH:    m->fp_mp = mp_koch; m->fp_gc = gc_koch; break;
    case mpMORI:    m->fp_mp = mp_mori; m->fp_gc = gc_mori; break;
    case mpKUEH:    m->fp_mp = mp_plac; m->fp_gc = gc_plac; break;
    case mpPOPA:    m->fp_mp = mp_plac; m->fp_gc = gc_plac; break;
    
    default:
      m->fp_mp = mp_meri;    /* set meridian system */
      m->fp_gc = gc_meri;
      return TRUE;
  }
  return FALSE;
}

/* some precalculations
//----------------------------------------------------------*/
static int initVars(MUNPOS *m)
{
  m->eps = m->i_eps * RAD;
  m->lst = (m->i_st + m->i_lon) * RAD;
  m->phi = m->flags & mpPOLAR ? m->i_phi * RAD - ARC90 : m->i_phi * RAD;
  m->phi = normARC90(m->phi);
  m->shift = m->i_shift * RAD;
  
  m->mc = gc2lon(m->eps, m->lst, 0.);
  m->ac = gc2lon(m->eps, m->lst + ARC90, m->phi);
  m->flags = norm2PI(m->ac - m->mc) < PI ? m->flags & ~mpSTA_DIR : m->flags | mpSTA_DIR;
  if(m->flags & mpSTA_DIR) 
    m->ac += PI;
  m->flags &= ~(mpERRORS);  /* reset errors RCP and DCP */

  m->sin_eps = sin(m->eps);
  m->tan_phi = tan(m->phi);

  switch(m->hsys)
  {
    case mpPOLI:
    case mpALCA:
      m->raac = atan2(cos(m->eps) * sin(m->ac), cos(m->ac));
      break;

    case mpHALY:
      m->de = asin(sin(m->ac) * sin(m->eps));
      m->ad = asin(tan(m->de) * tan(m->phi));
      break;

    case mpPLAC:
    case mpKUEH:
      if(fabs(m->phi) >= ARC90 - m->eps - TINY_RAD)
        m->flags |= mpERR_RCP;  /* circumpoler region */
      break;

    case mpKOCH:
      if(fabs(m->phi) >= ARC90 - m->eps - TINY_RAD)
      {
        m->flags |= mpERR_RCP;  /* circumpoler region */
        break;
      }
      m->de = asin(sin(m->mc) * sin(m->eps));
      if(fabs(m->de) >= ARC90 - fabs(m->phi))
        m->flags |= mpERR_DCP;  /* declination circle in circumpoler region */
      else
        m->ad = asin(tan(m->de) * tan(m->phi));
      break;
  }
  return m->flags & mpERRORS ? TRUE : FALSE;
}


/*============================================================
// public
//==========================================================*/

/* constructor
//----------------------------------------------------------*/
void mpInit(MUNPOS *m)
{
  m->i_eps = 24.0;
  m->i_st = 0.0;
  m->i_lon = 0.0;
  m->i_phi = 45.0;
  mpSetHsys(m, mpMERI);
}


/*------------------------------------------------------------
// setter functions
//----------------------------------------------------------*/

/* sets the house system
//----------------------------------------------------------*/
int mpSetHsys(MUNPOS *m, unsigned hsys)
{
  initFP(m, hsys);
  m->i_shift = 0.0;
  return initVars(m);
}

/* sets the house-system and the mundane shift
//----------------------------------------------------------*/
int mpSetHsysShift(MUNPOS *m, unsigned hsys, double shift)
{
  initFP(m, hsys);
  m->i_shift = shift;
  return initVars(m);
}

/* sets the mundane shift
//----------------------------------------------------------*/
void mpSetShift(MUNPOS *m, double shift)
{
  m->i_shift = shift;
  m->shift = shift * RAD;
}

/* sets the declination of parallel circle systems
//----------------------------------------------------------*/
int mpSetDecl(MUNPOS *m, double de)
{
  if(m->hsys != mpHALY)
    return TRUE;    /* invalid house system */
  m->de = de * RAD;
  if(fabs(m->de) >= ARC90 - fabs(m->phi)) 
  {
    m->flags |= mpERR_DCP;
    return TRUE;    /* declination circle in circumpolar region */
  }
  m->ad = asin(tan(m->de) * tan(m->phi));
  return FALSE;
}

/* sets the the event time and location
//----------------------------------------------------------*/
void mpSetEvent(MUNPOS *m, double eps, double st, double lon, double lat)
{
  m->i_eps = eps;
  m->i_st = st;
  m->i_lon = lon;
  m->i_phi = lat;
  initVars(m);
}

/* sets the event location
//----------------------------------------------------------*/
void mpSetLocation(MUNPOS *m, double lon, double lat)
{
  m->i_lon = lon;
  m->i_phi = lat;
  initVars(m);
}

/* sets the event time
//----------------------------------------------------------*/
void mpSetTime(MUNPOS *m, double eps, double st)
{
  m->i_eps = eps;
  m->i_st = st;
  initVars(m);
}

/* sets the oblique of the ecliptic
//----------------------------------------------------------*/
void mpSetEPS(MUNPOS *m, double eps)
{
  m->i_eps = eps;
  initVars(m);
}

/* sets the apparent sideral time (ramc)
//----------------------------------------------------------*/
void mpSetST(MUNPOS *m, double st)
{
  m->i_st = st;
  initVars(m);
}

/* sets the geographic longitude
//----------------------------------------------------------*/
void mpSetLon(MUNPOS *m, double lon)
{
  m->i_lon = lon;
  initVars(m);
}

/* sets the geographic latitude
//----------------------------------------------------------*/
void mpSetLat(MUNPOS *m, double lat)
{
  m->i_phi = lat;
  initVars(m);
}


/*------------------------------------------------------------
// getter functions
//----------------------------------------------------------*/

/* gets the house system
//----------------------------------------------------------*/
unsigned mpGetHsys(MUNPOS *m)
{
  return m->hsys;
}

/* gets the flags
//----------------------------------------------------------*/
unsigned mpGetFlags(MUNPOS *m)
{
  return m->flags;
}

/* gets the shift value
//----------------------------------------------------------*/
double mpGetShift(MUNPOS *m)
{
  return m->i_shift;
}

/* gets the oblique of the ecliptic
//----------------------------------------------------------*/
double mpGetEPS(MUNPOS *m)
{
  return m->i_eps;
}

/* gets the apparent sideral time
//----------------------------------------------------------*/
double mpGetST(MUNPOS *m)
{
  return m->i_st;
}

/* gets the longitude of the event location
//----------------------------------------------------------*/
double mpGetLon(MUNPOS *m)
{
  return m->i_lon;
}

/* gets the latitude of the event location
//----------------------------------------------------------*/
double mpGetLat(MUNPOS *m)
{
  return m->i_phi;
}

/* gets the apparent local sideral time (RAMC)
//----------------------------------------------------------*/
double mpGetRAMC(MUNPOS *m)
{
  return m->lst / RAD;
}

/* medium coeli
//----------------------------------------------------------*/
double mpGetMC(MUNPOS *m)
{
  return m->mc / RAD;
}

/* ascendant
//----------------------------------------------------------*/
double mpGetAC(MUNPOS *m)
{
  double psi = m->flags & mpPOLAR ? m->phi + ARC90 : m->phi;
  double ac = gc2lon(m->eps, m->lst + ARC90, psi);
  if(norm2PI(ac - m->mc) >= PI)
    ac = norm2PI(ac + PI);
  return ac / RAD;
}

/* antivertex
//----------------------------------------------------------*/
double mpGetAV(MUNPOS *m)
{
  double psi = m->flags & mpPOLAR ? m->phi : m->phi - ARC90;
  double ax = gc2lon(m->eps, m->lst + ARC90, psi);
  if(norm2PI(ax - m->mc) >= PI)
    ax = norm2PI(ax + PI);
  return ax / RAD;
}

/* equatorial ascendant, polar antivertex
//----------------------------------------------------------*/
double mpGetEA(MUNPOS *m)
{
  return gc2lon(m->eps, m->lst + ARC90, 0.0) / RAD;
}


/*------------------------------------------------------------
// calculations
//----------------------------------------------------------*/

/* calculates n house cusps
//----------------------------------------------------------*/
int mpCalcCusps(MUNPOS *m, int n, double *cusps)
{
  double mp, width;
  int i;
  mp = 0.0;
  width = 360.0 / (double)n;
  for(i = 0; i < n; i++) 
  {
    cusps[i] = mpCalcLong(m, mp);
    mp += width;
  }
  return m->flags & mpERRORS ? TRUE : FALSE;
}

/* longitude from mundane position mp
//----------------------------------------------------------*/
double mpCalcLong(MUNPOS *m, double mp)
{
  double lon, gc[2];
  if(m->flags & mpERRORS)
    return mpERROR;
  mp *= RAD;
  /* apparent to true mundan position */
  mp = m->flags & mpREVERSE ? -(mp + m->shift) : mp + m->shift;
  m->fp_gc(m, mp, gc);
  lon = gc2lon(m->eps, gc[0], gc[1]);
  if((m->flags & (mpSTA_DIR | mpSWAP)) == (mpSTA_DIR | mpSWAP))
    lon += PI;
  return norm2PI(lon) / RAD;
}

/* longitude from OA and pole
//----------------------------------------------------------*/
double mpCalcLong2(MUNPOS *m, double oa, double psi)
{
  double lon;
  lon = gc2lon(m->eps, oa * RAD, psi * RAD) / RAD;
  if((m->flags & (mpSTA_DIR | mpSWAP)) == (mpSTA_DIR | mpSWAP))
    lon += 180.0;
  return mpNorm360(lon);
}

/* directed longitude
//----------------------------------------------------------*/
double mpCalcLongDir(MUNPOS *m, double lon, double lat, double dir_arc)
{
  double ramc, mp, ra, de, pd, psi, ad, ol, gc[2];
  lon *= RAD; lat *= RAD; dir_arc *= RAD;
  ra = ecl2ra(m->eps, lon, lat);
  de = ecl2de(m->eps, lon, lat);

  if((mp = m->fp_mp(m, ra, de)) == mpERROR) 
    return mpERROR;
      
  switch(m->hsys)
  {
    case mpPLAC:
      ramc = m->lst;
      m->lst += dir_arc;
      gc_plac(m, mp, gc);
      ol = gc2lon(m->eps, gc[0], gc[1]);
      m->lst = ramc;
      break;

    case mpKUEH:
      psi = atan(sin(ra - mp - ARC90 - m->lst) / tan(de));
      ol = gc2lon(m->eps, m->lst + ARC90 + mp + dir_arc, psi);
      break;

    case mpPOPA:
      if((mp = mp_plac(m, ra, de)) == mpERROR)
        return mpERROR;
      pd = PF(mp);
      psi = atan(pd * tan(m->phi));
      ad = asin(tan(de) * tan(psi));
      ol = gc2lon(m->eps, ra - ad + dir_arc, psi);
      break;

    default:
      m->fp_gc(m, mp, gc);
      ol = gc2lon(m->eps, gc[0] + dir_arc, gc[1]);
      break;
  }
  return ol / RAD;
}


/* mundane position from longitude and latitude
//----------------------------------------------------------*/
double mpCalcPosition(MUNPOS *m, double lon, double lat)
{
  return mpCalcPos2(m, lon, lat, TRUE);
}

/* mundane position
//----------------------------------------------------------*/
double mpCalcPos2(MUNPOS *m, double ra, double de, int ecl)
{
  double mp;
  ra *= RAD; de *= RAD;
  if(ecl) 
  {
    mp = ecl2ra(m->eps, ra, de);
    de = ecl2de(m->eps, ra, de);
    ra = mp;
  }
  if((mp = m->fp_mp(m, ra, de)) == mpERROR) 
    return mpERROR;
  else
    return norm2PI(m->flags & mpREVERSE ? -mp - m->shift : mp - m->shift) / RAD;
}

/* falified mundane position
//----------------------------------------------------------*/
double mpCalcPosIP(MUNPOS *m, int n, double lon)
{
  double seg, body_mp, cusp_mp, cusp_x, cusp_y, fact;
  if(m->flags & mpERRORS)
    return mpERROR;
  if(n < 4) n = 4;
  seg = 360.0 / (double)n;
  body_mp = mpCalcPosition(m, lon, 0.0);     
  cusp_mp = Trunc(body_mp / seg + TINY) * seg;
  cusp_x  = mpCalcLong(m, cusp_mp);
  cusp_y  = mpCalcLong(m, cusp_mp + seg);
  fact    = mpADif(cusp_x, lon) / mpADif(cusp_x, cusp_y);
  return cusp_mp + fact * seg;
}

/* direction to mundane Position
//----------------------------------------------------------*/

double mpCalcArc(MUNPOS *m, double pro_lon, double pro_lat, double mp, unsigned flags, double asp)
{
  double pro[2], dir_arc;
  pro_lon *= RAD; pro_lat *= RAD; mp *= RAD; asp *= RAD;
  mp = norm2PI(m->flags & mpREVERSE ? -(mp + m->shift) : mp + m->shift);

  if(!(flags & mpLAT_P)) pro_lat = 0.0;
  if(flags & mpBIANC) pro_lat = asin(sin(pro_lat) * cos(asp));

  if(flags & mpMUNDO)  /* in mundo */
    mp = flags & mpMIRROR ? asp + asp - mp : mp + asp;
  else
    pro_lon += asp;    /* in zodiaco */     

  cotrans(m->eps, pro_lon, pro_lat, pro);
  dir_arc = dir1(m, pro, mp);
  return dir_arc == mpERROR ? mpERROR : dir_arc / RAD;
}

/* interplanatary direction
//----------------------------------------------------------*/
double mpCalcArc2(MUNPOS *m, double pro_lon, double pro_lat, double sig_lon, double sig_lat, unsigned flags, double asp)
{
  double pro[2], sig[2], mp, dir_arc;
  pro_lon *= RAD; pro_lat *= RAD; sig_lon *= RAD; sig_lat *= RAD; asp *= RAD;
  
  if(!(flags & mpLAT_S)) sig_lat = 0.0;
  if(!(flags & mpLAT_P)) pro_lat = 0.0;
  if(flags & mpBIANC) pro_lat = asin(sin(pro_lat) * cos(asp));

  if(flags & mpMUNDO) /* in mundo */
  {
    cotrans(m->eps, pro_lon, pro_lat, pro);
    cotrans(m->eps, sig_lon, sig_lat, sig);
    if((mp = m->fp_mp(m, sig[0], sig[1])) == mpERROR) 
      return mpERROR;
    mp = flags & mpMIRROR ? asp + asp - mp : mp + asp;
    dir_arc = dir1(m, pro, mp);
  }
  else               /* in zodiaco */
  {
    if(flags & mpASP2SIG)
      sig_lon += asp;
    else
      pro_lon += asp;
    cotrans(m->eps, pro_lon, pro_lat, pro);
    cotrans(m->eps, sig_lon, sig_lat, sig);
    dir_arc = dir2(m, pro, sig);
  }
  return dir_arc == mpERROR ? mpERROR : dir_arc / RAD;
}

int mpCalcArcRapt(MUNPOS *m, double lon1, double lat1, double lon2, double lat2, double mpos, double *vect)
{
  double v1[2], v2[2];
  int it1, it2, it;
  if((it1 = mpCalcRapt(m, lon1, lat1, lon2, lat2, mpos, v1)) < 0)
    return -1;
  if((it2 = mpCalcRapt(m, lon1, lat1, lon2, lat2, mpos + 180.0, v2)) < 0)
    return -1;
  it = it1 + it2;
  /* conjunction */
  if(fabs(v1[1]) < 180.0 - v2[1]) {
    vect[0] = v1[0];    
    vect[1] = v1[1];
  }    
  else {
    vect[0] = v2[0];
    vect[1] = 180.0 - v2[1];
  }
  /* opposition */
  if(fabs(v2[1]) < 180.0 - v1[1]) {
    vect[2] = v2[0];
    vect[3] = v2[1];
  }
  else {
    vect[2] = v1[0];
    vect[3] = 180.0 - v1[1];
  }
  /* smaller equidistance */    
  if(fabs(vect[1]) < fabs(vect[3])) {
    vect[4] = vect[0];
    vect[5] = vect[1];
  }
  else {
    vect[4] = vect[2];
    vect[5] = vect[3];
  }
  return it;
}

/* rapt parallel
//----------------------------------------------------------*/
int mpCalcRapt(MUNPOS *m, double lon1, double lat1, double lon2, double lat2, double mp, double *vect)
{
  double start, delta, de1, de2, mx, MP1, MP2, mp1, mp2, mp3;
  double x1, x2, x3, y1, y2, y3, yshift;
  const double iwidth = ARC360 - TINY_RAD;
  int i;
  lon1 *= RAD; lat1 *= RAD; lon2 *= RAD; lat2 *= RAD, mp *= RAD;  
  mp = norm2PI(m->flags & mpREVERSE ? -(mp + m->shift) : mp + m->shift);

  start = ecl2ra(m->eps, lon1, lat1);
  delta = ecl2ra(m->eps, lon2, lat2) - start;

  de1 = ecl2de(m->eps, lon1, lat1);
  de2 = ecl2de(m->eps, lon2, lat2);

  if((MP1 = m->fp_mp(m, start, de1)) == mpERROR)
    return -1;
  if((MP2 = m->fp_mp(m, start + delta + iwidth, de2)) == mpERROR)
    return -1;
  MP1 = norm2PI(MP1);
  MP2 = norm2PI(MP2);
  
  x1 = start;                       /* ra1 */
  x2 = start + iwidth;              /* ra2 */

  y1 = (MP1 + MP2) / 2.0 - mp;      /* mp1 */
  y2 = y1 + iwidth;                 /* mp2 */

  yshift = (y1 * y2 > 0) ? ARC360 : 0.;
  y1 -= yshift;
  y2 -= yshift;

  for(i = 0; i < 32; i++)
  {
    mx = (y2 - y1) / (x2 - x1);
    x3 = x1 - y1 / mx;

    if((mp1 = m->fp_mp(m, x3, de1)) == mpERROR)
      return -1;
    if((mp2 = m->fp_mp(m, x3 + delta, de2)) == mpERROR)
      return -1;
    mp1 = norm2PI(mp1);
    mp2 = norm2PI(mp2);
    if(mp1 < MP1) mp1 += ARC360;
    if(mp2 < MP2) mp2 += ARC360;

    mp3 = (mp1 + mp2) / 2.0;     
    if(fabs(mp - mp3 + yshift) < 0.0000005 * RAD)
       break;
    
    y3 = mp3 - mp - yshift;
    if(y3 * y2 > 0.0) 
    {
      x1 = x2;
      y1 = y2;
    }
    else
      y1 *= y2 / (y2 + y3);   /* Pegasus */

    x2 = x3;
    y2 = y3;
  }
  vect[0] = normPI(start - x3) / RAD;
  vect[1] = norm2PI(mp - mp1) / RAD;
  if(i >= 32) 
    i = -1;
  return i;
}


/* rectascension from longitude and latitude
//----------------------------------------------------------*/
double mpCalcRA(MUNPOS *m, double lon, double lat)
{
  return ecl2ra(m->eps, lon * RAD, lat * RAD) / RAD;
}

/* declination from longitude and latitude
//----------------------------------------------------------*/
double mpCalcDE(MUNPOS *m, double lon, double lat)
{
  return ecl2de(m->eps, lon * RAD, lat * RAD) / RAD;
}

/* azimuth from longitude and latitude
//----------------------------------------------------------*/
double mpCalcAZ(MUNPOS *m, double lon, double lat)
{
  double ra = ecl2ra(m->eps, lon * RAD, lat * RAD);
  double de = ecl2de(m->eps, lon * RAD, lat * RAD);
  double ha = m->lst - ra;
  double az = atan2(sin(ha), cos(ha) * sin(m->i_phi * RAD) - tan(de) * cos(m->i_phi * RAD));
  az = norm2PI(az + PI) / RAD;
  if(az > 360.0 - TINY) az = 0.0;
  return az;
}

/* altitude from longitude and latitude
//----------------------------------------------------------*/
double mpCalcALT(MUNPOS *m, double lon, double lat)
{
  double ra = ecl2ra(m->eps, lon * RAD, lat * RAD);
  double de = ecl2de(m->eps, lon * RAD, lat * RAD);
  double ha = m->lst - ra;
  return asin(sin(de) * sin(m->i_phi * RAD) + cos(de) * cos(m->i_phi * RAD) * cos(ha)) / RAD;
}

/* Upper Meridian Distance
//----------------------------------------------------------*/
double mpCalcUMD(MUNPOS *m, double ra)
{
  return mpNorm180(ra - m->lst / RAD);
}

/* Ascensional- / Azimuthaldifference
//----------------------------------------------------------*/
double mpCalcAD(MUNPOS *m, double de)
{
  double sin_ad = tan(m->phi) * tan(de * RAD);
  if(fabs(sin_ad) > 1.0)
    return mpERROR;
  else
    return asin(sin_ad) / RAD;
}

/* Semiarc
//----------------------------------------------------------*/
double mpCalcSA(MUNPOS *m, double ra, double de)
{
  double umd = fabs(mpNorm180(ra - m->lst / RAD));
  double ad = mpCalcAD(m, de);
  if(ad == mpERROR)
    return mpERROR;
  else
    return fabs(umd > 90.0 + ad ? -(90.0 - ad) : 90.0 + ad);
}

/* OA or OD from longitude, latitude and pole
//----------------------------------------------------------*/
double mpCalcOA(MUNPOS *m, double ra, double de, int flag)
{
  double ad = mpCalcAD(m, de);
  if(ad == mpERROR)
    return ad;
  else
    return flag ? (ra + ad)  : (ra - ad);
}

/* OA and pole from mundane position mp
//----------------------------------------------------------*/
int mpCalcGC(MUNPOS *m, double mp, double *gc)
{
  if((m->flags & mpERRORS) || m->hsys == mpKUEH || m->hsys == mpPOPA)
    return TRUE;
  mp *= RAD;
  /* apparent to true mundan position */
  mp = m->flags & mpREVERSE ? -(mp + m->shift) : mp + m->shift;
  m->fp_gc(m, mp, gc);
  gc[0] /= RAD;
  gc[1] /= RAD;
  return FALSE;
}    

double mpCalcGC2(MUNPOS *m, double ra, double de, double *gc)
{
  double mp;
  if((mp = gc2(m, ra * RAD, de * RAD, gc)) == mpERROR)
    return mpERROR;
  gc[0] /= RAD;
  gc[1] = tan(gc[1]) / RAD;
  return mp / RAD;
}

/* various data for the speculum
//----------------------------------------------------------*/
void mpCalcSpeculum(MUNPOS *m, double lon, double lat, SPECULUM *spec)
{
  double ra, de, phi, gc[2], pd[6], umd, mp;
  ra  = ecl2ra(m->eps, lon * RAD, lat * RAD);
  de  = ecl2de(m->eps, lon * RAD, lat * RAD);
  phi = m->i_phi * RAD;
  umd = normPI(ra - m->lst); 
    
  spec->lon = lon;
  spec->lat = lat;
  spec->ra  = ra / RAD;
  spec->de  = de / RAD;
  spec->az  = norm2PI((-atan2(sin(umd), cos(umd) * sin(phi) - tan(de) * cos(phi))) + PI) / RAD;
  spec->alt = asin(sin(de) * sin(phi) + cos(de) * cos(phi) * cos(-umd)) / RAD;

  spec->md = fabs(umd) < ARC90 ? fabs(umd) / RAD : -(PI - fabs(umd)) / RAD;  
  if(!pdist(m, ra, de, pd))  /* ad ok? */
  {
    spec->ad = pd[4] / RAD;
    spec->sa = fabs(pd[0]) > pd[2] ? -pd[3] / RAD : pd[2] / RAD;
    spec->pd = pd[5];
    spec->oa = pd[0] < 0.0 ? (ra + pd[4]) / RAD : (ra - pd[4]) / RAD;
    spec->hd = pd[0] < 0.0 ? -fabs(spec->sa - spec->md) : fabs(spec->sa - spec->md);
  }
  else
    spec->ad = spec->sa = spec->pd = spec->oa = spec->hd = mpERROR;

  /* calc mp, OA/OD and tan(pole) */
  mp = gc2(m, ra, de, gc);
  if(mp != mpERROR) 
  {
    /* apparent mundane position */
    spec->mp = norm2PI(m->flags & mpREVERSE ? -mp - m->shift : mp - m->shift) / RAD;
    /* derived from mp */
    spec->w = norm2PI(gc[0]) / RAD;
    spec->psi = atan(gc[1]) / RAD;
    spec->q = asin(fabs((gc[1])) * tan(de)) / RAD;
  }
  else
    spec->mp = spec->q = spec->w = spec->psi = mpERROR;

  /* zenith distance */
  spec->zd = fabs(normPI(ecl2ra(m->i_phi * RAD, ra - ARC90 - m->lst, de) - ARC90)) / RAD;
  spec->zd = spec->zd < 90.0 ? -spec->zd : 180.0 - spec->zd;  
  /* east point distance */
  spec->epd = fabs(normPI(ecl2ra(m->i_phi * RAD - ARC90, ra - ARC90 - m->lst, de))) / RAD;
  spec->epd = spec->epd < 90.0 ? spec->epd : -(180.0 - spec->epd);  
  /* right longitude */
  spec->rl = atan2(sin(ra), cos(ra) * cos(m->eps)) / RAD;
  if(spec->rl < 0.0) 
    spec->rl += 360.0;
  /*
  if(fabs(umd) > dsa)
    q = umd >= 0.0 ? 0 : 1;
  else
    q = umd < 0.0  ? 2 : 3;
  */
}


/*------------------------------------------------------------
// conversions
//----------------------------------------------------------*/

/* Ecliptic <-> Equator
//----------------------------------------------------------*/
void mpCoTrans(double eps, double lon, double lat, double *vect)
{
  cotrans(eps * RAD, lon * RAD, lat * RAD, vect);
  vect[0] /= RAD;
  vect[1] /= RAD;
}

/* Equator -> Horizon
//----------------------------------------------------------*/
void mpEqu2Hor(double lst, double phi, double ra, double de, double *hor)
{
  lst *= RAD; phi *= RAD;  ra *= RAD; de *= RAD;
  hor[0] = equ2az(lst, phi, ra, de) / RAD;
  hor[1] = equ2alt(lst, phi, ra, de) / RAD;
}

/* Horizon -> Equator
//----------------------------------------------------------*/
void mpHor2Equ(double lst, double phi, double az, double alt, double *equ)
{
  lst *= RAD; phi *= RAD; az = (az - 180) * RAD; alt *= RAD;
  equ[0] = (lst - atan2(sin(az), cos(az) * sin(phi) + tan(alt) * cos(phi))) / RAD;
  equ[1] = asin(sin(phi) * sin(alt) - cos(phi) * cos(alt) * cos(az)) / RAD;
}

/* shift to range 0..360 deg
//----------------------------------------------------------*/
double mpNorm360(double x)
{
  x = fmod(x, 360.);
  if(x < 0)
    x += 360.;
  return(x);
}

/* shift to range -180..+180 deg
//----------------------------------------------------------*/
double mpNorm180(double x)
{
  x = fmod(x, 360.);
  if(x < 0)
    x += 360.;
  if(x >= 180.)
    x -= 360.;
  return(x);
}

/* Julian day from year, month, day and hours
//----------------------------------------------------------*/
double mpCal2JD(int year, int month, int day, double hours, int julian, DATTIM *m)
{
/*  int a, b; */
  double a, b, c, jd;
  c = ((double)day) + hours / 24.0 - 1524.5;
  if(month < 3) 
  {
    year--;
    month += 12;
  }
  b = 0.0;
  if(!julian) 
  {
    a = Trunc(year / 100.0);
    b = 2 - a + Trunc(a / 4.0);
  }
  jd = Trunc(365.25 * (year + 4716) + Trunc(30.6001 * (month + 1)) + b) + c;
  if(m != NULL)
  {
    m->year = year;
    m->month = month;
    m->day = day;
    m->hours = hours;
    m->jd = jd;
  }
  return jd;
}

/* Year, month, day and hours from Julian day
//----------------------------------------------------------*/
void mpJD2Cal(double jd, DATTIM *m)
{
  double a, b, c, d, e, z, alf;
  m->jd = jd;
  jd += 0.5;
  z = Trunc(jd);
  m->hours = (jd - z) * 24.0;
  if(z < 2299161.0)
    a = z;
  else {
    alf = Trunc((z - 1867216.25) / 36524.25);
    a = z + 1 + alf - Trunc(alf / 4.0);
  }
  b = a + 1524;
  c = Trunc((b - 122.1) / 365.25);
  d = Trunc(365.25 * c);
  e = Trunc((b - d) / 30.6001);
  m->day = (int)(b - d - Trunc(30.6001 * e));
  if (e < 14)
    m->month = (int)(e - 1);
  else
    m->month = (int)(e - 13);
  if (m->month > 2)
    m->year = (int)(c - 4716);
  else
    m->year = (int)(c - 4715);
}


/*------------------------------------------------------------
// some additional stuff
//----------------------------------------------------------*/

/* difference a, b (clockwise negative)
//----------------------------------------------------------*/
double mpDif(double a, double b)
{
  return mpNorm180(b - a);
}

/* difference a, b (always positive)
//----------------------------------------------------------*/
double mpADif(double a, double b)
{
  return fabs(mpNorm180(b - a));
}

/* arithmetic midpoint
//----------------------------------------------------------*/
double mpMid(double a, double b)
{
  double x = mpNorm180(a - b) / 2.;
  x += b;
  return mpNorm360(x);
}

/* Orthodrome, geometric midpoint
//----------------------------------------------------------*/
double mpAspect(double Alon, double Aphi, double Blon, double Bphi, double *vect)
{
  double dist, gc[2];
  Alon *= RAD; Aphi *= RAD; Blon *= RAD; Bphi *= RAD;
  dist = acos(sin(Aphi) * sin(Bphi) + cos(Aphi) * cos(Bphi) * cos(Blon - Alon));
  if(vect != NULL)
  {
    cotrans(Aphi, Blon - Alon - ARC90, Bphi, gc);
    cotrans(atan2(tan(gc[1]), sin(gc[0] + ARC90)), dist / 2.0, 0.0, gc);
    cotrans(-Aphi, gc[0] - ARC90, gc[1], gc);
    vect[0] = norm2PI(gc[0] + ARC90 + Alon) / RAD;
    vect[1] = normARC90(gc[1]) / RAD;
  }
  return dist / RAD;
}

/* transform gepraphic to geocentric latitude
//----------------------------------------------------------*/
double mpGeoG2Z(double phi)
{
  phi = atan((1.0 - pow(0.08181922, 2.0)) * tan(phi * RAD)) / RAD;
  return phi;  
}

/* Ascensionaldifference from Pole and Deklination
//----------------------------------------------------------*/
double mpAD(double psi, double de)
{
  double ad;
  if(fabs(ad = tan(psi * RAD) * tan(de * RAD)) > 1.0)
    return mpERROR;
  return asin(ad) / RAD;
}

/* OA or OD from rectascension, declination and pole
//----------------------------------------------------------*/
double mpOA(MUNPOS *m, double ra, double de, double psi, int flag)
{
  double ad = mpAD(psi, de);
  if(ad == mpERROR)
    return mpERROR;
  else
    return flag ? ra + ad : ra - ad;
}

/* ecl. longitude from eps, OA and pole
//----------------------------------------------------------*/
double mpLongGC(double eps, double oa, double psi)
{
  return gc2lon(eps * RAD, oa * RAD, psi * RAD) / RAD;
}



