/***
* munpos.h
*
* Copyright (c) 2015, Wolfgang Astl. All rights reserved.
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
* MA 02110-1301 USA
*
*/

/* Created   2005/11/13 */
/* Last Mod. 2015/12/14 */


#ifndef _MUNPOS_H_
#define	_MUNPOS_H_

#define mpSWAP        0x40    /* swap opposite cusps */
#define mpPOLAR       0x80    /* azimuthal */
#define mpREVERSE    0x100    /* clockwise */

/* house systems */
#define mpMCEQ    0           /* MC-equal */
#define mpACEQ    1           /* AC-equal */
#define mpPORP    2           /* Porphyry */
#define mpMERI    3           /* Meridian */
#define mpPOLI    4           /* Poli-Equatorial */
#define mpALCA    5           /* Alcabitus */
#define mpCAMP   (6 | mpSWAP) /* Campanus */
#define mpREGI   (7 | mpSWAP) /* Regiomontanus */
#define mpHALY   (8 | mpSWAP) /* Abenragel */
#define mpPLAC    9           /* Placidus */
#define mpTOPO (0xa | mpSWAP) /* Topocentric */
#define mpMORI  0xb           /* Morinus */
#define mpKOCH  0xc           /* Koch (GOHS) */

#define mpKUEH (0xd | mpSWAP) /* Kuehr */
#define mpPOPA (0xe | mpSWAP) /* Polich & Page */

/* flags */
#define mpERR_RCP     0x1000    /* circumpolar region */
#define mpERR_DCP     0x2000    /* declination circle in circimpolar region */
#define mpERRORS (mpERR_RCP | mpERR_DCP)

#define mpSTA_DIR     0x4000    /* AC or ATVX on wrong side */

#define mpERROR  -999.0

/* flags (mpCalcArc) */
#define mpLAT_S           0x2
#define mpLAT_P           0x4
#define mpLAT (LAT_S | LAT_P)
#define mpBIANC           0x8
#define mpASP2SIG        0x20
#define mpMUNDO       0x10000
#define mpMIRROR     0x100000

/* mpCalcDirList */
#define mpZOD             0x1
#define mpASP2PRO        0x10
#define mpRAPT         0x2000
#define mpCUSPS        0x4000
#define mpPAC           0x100
#define mpPMC           0x200
#define mpPMP           0x400
#define mpSAC        0x100000
#define mpSMC        0x200000
#define mpSMP        0x400000


#define mpRAD        0,017453292519943295769236907684886 
#define mpRad2Deg(x) (double) ((x) / mpRAD)
#define mpDeg2Rad(x) (double) ((x) * mpRAD)

typedef struct MUNPOS 
{
  unsigned hsys, flags;
  double i_shift;
  double i_eps, i_st, i_lon, i_phi;
  /* precalculated by initVars() */
  double shift, eps, lst, phi, mc, ac;
  double raac, de, ad, sin_eps, tan_eps, tan_phi;
  /* functors */
  double (*fp_mp) (struct MUNPOS*, double, double);
  void   (*fp_gc) (struct MUNPOS*, double, double*);
} MUNPOS;

typedef struct
{
  double lon;   /* ecl. Longitude */
  double lat;   /* ecl Latitude */
  double ra;    /* rectascension */
  double de;    /* declination */
  double az;    /* azimut */
  double alt;   /* altitude */

  double mp;    /* mundane position */
  double q;     /* AD (under the own pole) */
  double w;     /* OA/OD (under the own pole) */
  double psi;   /* pole */

  double umd;   /* upper meridian distance */
  double ad;    /* AD */
  double sa;    /* Semiarc */
  double oa;    /* OA/OD */

  double md;    /* meridian distance */
  double pd;    /* proportional distance */
  double hd;    /* horizon distance */
  double zd;    /* zenith distance */
  double epd;   /* eastpoint distance */
  double rl;    /* right longitude */
} SPECULUM;

typedef struct
{
  int year;
  int month;
  int day;
  double hours;
  double jd;
} DATTIM;

typedef struct {
  double arc;
  unsigned char pro;
  unsigned char sig;
  unsigned char asp;
  unsigned char stat;
} DIRLST;


#ifdef __cplusplus
extern "C" {
#endif

/**-----------------------------------------------------------
// Konstruktor
//----------------------------------------------------------*/

/**
 === Parameterloser Konstruktor ===
 Initialisiert die MUNPOS Instanz
   @param  *m   Pointer auf MUNPOS Instanz
*/
void mpInit(MUNPOS *m);


/**-----------------------------------------------------------
// Setter Funktionen
//----------------------------------------------------------*/

/**
 === Setzt das Haeusersystem ===
 Die Strukturvariable *shift* wird dabei auf 0 gesetzt.
   @param  *m        Pointer auf MUNPOS Instanz
   @param  hsys      Haeusersystem
   @flags  mpPOLAR   Azimutale Version berechnen
   @flags  mpREVERSE Haeuser im Uhrzeigersinn
   @return 0 oder im Fehlerfall 1
*/
int mpSetHsys(MUNPOS *m, unsigned hsys);

/**
 === Setzt das Haeusersystem und den mundanen Offset *shift* ===
   @param  *m        Pointer auf MUNPOS Instanz
   @param  hsys      Haeusersystem
   @param  shift     Offset der mundanen Position (ueblicherweise 0.0)
   @flags  mpPOLAR   Azimutale Version berechnen
   @flags  mpREVERSE Haeuser im Uhrzeigersinn
   @return 0 oder im Fehlerfall 1
*/
int mpSetHsysShift(MUNPOS *m, unsigned hsys, double shift);

/**
 ===  Setzt den mundanen Offset *shift* ===
   @param *m    Pointer auf MUNPOS Instanz
   @param shift Offset der mundanen Position
*/
void mpSetShift(MUNPOS *m, double shift);

/**
 === Setzt die Deklination fuer Parallelkreissysteme ===
 Setzt die Deklination fuer Deklinationssysteme (z.B. Solar House 
 System). *Zuvor* muss das Haeusersysten mpHALY gesetzt werden.
   @param  *m Pointer auf MUNPOS Instanz
   @param  de Deklination des Parallelkreises
   @return 0 oder im Fehlerfall 1
*/
int mpSetDecl(MUNPOS *m, double de);

/**
 === Setzt Zeitpunkt und Ort des Ereignisses ===
 Setzt die Schiefe der Ekliptik, die Ereigniszeit und den Ereignisort.
   @param *m    Pointer auf MUNPOS Instanz
   @param eps   Wahre Schiefe der Ekliptik
   @param st    Sternzeit (Greenwich apparent sidereal time)
   @param lon   Geographische Laenge (positiv nach Osten gezaehlt)
   @param lat   Geographische Breite (auf der Nordhalbkugel positiv, auf der Suedhalbkugel negativ)
*/
void mpSetEvent(MUNPOS *m, double eps, double st, double lon, double lat);

/**
 === Setzt die geographische Laenge und Breite des Ereignisortes ===
   @param *m   Pointer auf MUNPOS Instanz
   @param lon  Geographische Laenge (positiv nach Osten gezaehlt)
   @param lat  Geographische Breite (auf der Nordhalbkugel positiv, auf der Suedhalbkugel negativ)
*/
void mpSetLocation(MUNPOS *m, double lon, double lat);

/**
 === Setzt den Zeitpunkt des Ereignises ===
   @param *m  Pointer auf MUNPOS Instanz
   @param eps Wahre Schiefe der Ekliptik
   @param st  Sternzeit (Greenwich apparent sidereal time)
*/
void mpSetTime(MUNPOS *m, double lon, double lat);

/**
 === Setzt die Schiefe der Ekliptik. ===
   @param *m    Pointer auf MUNPOS Instanz
   @param eps   Wahre Schiefe der Ekliptik
*/
void mpSetEPS(MUNPOS *m, double eps);

/**
 === Setzt die Sternzeit ===
   @param *m   Pointer auf MUNPOS Instanz
   @param st   Sternzeit (Greenwich apparent sidereal time)
*/
void mpSetST(MUNPOS *m, double st);

/**
 === Setzt die geographische Laenge ===
   @param *m   Pointer auf MUNPOS Instanz
   @param lon  Geographische Laenge (positiv nach Osten gezaehlt)
*/
void mpSetLon(MUNPOS *m, double lon);

/**
 === Setzt die geographische Breite ===
   @param *m   Pointer auf MUNPOS Instanz
   @param lat  Geographische Breite (auf der Nordhalbkugel positiv, auf der Suedhalbkugel negativ)
*/
void mpSetLat(MUNPOS *m, double ast);


/**----------------------------------------------------------
// Getter Funktionen
//----------------------------------------------------------*/

/**
 === Gibt das Haeusersystem zurueck ===
   @param  *m Pointer auf MUNPOS Instanz
   @return Haeusersystem
*/
unsigned mpGetHsys(MUNPOS *m);

/**
 === Gibt die Flags zurueck ===
   @param  *m        Pointer auf MUNPOS Instanz
   @flags  mpPOLAR   Azimutale Version berechnen
   @flags  mpREVERSE Haeuser im Uhrzeigersinn.
   @flags  mpSTA_DIR Aszendent oder Antivertex auf der 'falschen' Seite des Meridians.
   @return flags als unsigned int
*/
unsigned mpGetFlags(MUNPOS *m);

/**
 === Gibt der mundanen Offset zurueck ===
   @param  *m        Pointer auf MUNPOS Instanz
   @return Wert der Variablen shift.
*/
double mpGetShift(MUNPOS *m);

/**
 === Gibt die Schiefe der Ekliptik zurueck ===
   @param  *m        Pointer auf MUNPOS Instanz
   @return Wahre Schiefe der Ekliptik
*/
double mpGetEPS(MUNPOS *m);

/**
 === Gibt die Sternzeit zurueck ===
   @param  *m   Pointer auf MUNPOS Instanz
   @return Sternzeit (Greenwich apparent sideral time)
*/
double mpGetST(MUNPOS *m);

/**
 === Gibt die geographische Laenge des Ereignisortes zurueck ===
   @param  *m Pointer auf MUNPOS Instanz
   @return Geographische Laenge (positiv nach Osten gezaehlt)
*/
double mpGetLon(MUNPOS *m);

/**
 === Gibt die geographische Breite des Ereignisortes zurueck ===
   @param  *m Pointer auf MUNPOS Instanz
   @return Geographische Breite (auf der Nordhalbkugel positiv, auf der Suedhalbkugel negativ)
*/
double mpGetLat(MUNPOS *m);

/**
 === Gibt die lokale Sternzeit im Gradmass (RAMC) zurueck ===
   @param  *m Pointer auf MUNPOS Instanz
   @return Wahre lokale Sternzeit im Gradmass (RAMC)
*/
double mpGetRAMC(MUNPOS *m);

/**
 === Gibt die Laenge des MC zurueck ===
   @param  *m Pointer auf MUNPOS Instanz
   @return Laenge des MC
*/
double mpGetMC(MUNPOS *m);

/**
 === Gibt die Laenge des Aszendenten zurueck ===
   @param  *m Pointer auf MUNPOS Instanz
   @return Laenge des Aszendenten
*/
double   mpGetAC(MUNPOS *m);

/**
 === Berechnet die Laenge des Antivertex ===
 Dieser Punkt wird auch als *'Ostpunkt der Ekliptik'* bezeichnet.
   @param  *m Pointer auf MUNPOS Instanz
   @return Laenge des Antivertex
*/
double   mpGetAV(MUNPOS *m);

/**
 === Berechnet die Laenge des aequatorial-Aszendenten / Polar-Antivertex ===
 Dieser Punkt wird meistens faelschlicherweise als *'Ostpunkt'* bezeichnet.
   @param  *m Pointer auf MUNPOS Instanz
   @return Laenge des Aequatorial-Aszendenten / Polar-Antivertex
*/
double   mpGetEA(MUNPOS *m);


/**-----------------------------------------------------------
// Berechnungen
//----------------------------------------------------------*/

/**
 === Berechnet n Haeuserspitzen ===
 Schreibt die Laengen von n Haeusersptzen in das Array cusps[n].
   @param  *m    Pointer auf MUNPOS Instanz
   @param  n     Anzahl der Haeuser
   @param  cusps Pointer aud double[n]
   @return 0 oder im Fehlerfall 1
*/
int mpCalcCusps(MUNPOS *m, int n, double *cusps);

/**
 === Laenge aus mundaner Position ===
 Berechnet die Laenge aus der mundanen Position. Hat mundane Position die 
 Werte  0, 30, 60.., so entspricht diese Laenge einer Haeuserspitze eines
 12-teiligen Haeusersystems.
   @param  *m Pointer auf MUNPOS Instanz
   @param  mp mundane Position
   @return Laenge
*/
double mpCalcLong(MUNPOS *m, double mp);

/**
 === Laenge aus OA und Polhoehe ===
   @param  *m  Pointer auf MUNPOS Instanz
   @param  oa  Schiefe Aufsteigung
   @param  psi Polhoehe
   @return Laenge oder im Fehlerfall mpERROR
*/
double mpCalcLong2(MUNPOS *m, double oa, double psi);

/**
 === Dirigierte Position in schiefer Laenge ===
 Verschiebt die Positionslinie des durch Laenge und Breite gegebenen
 Signifikators unter Beibehaltung der Polhoehe um den Direktionsbogen
 entlang des aequators und gibt die Laenge des Schnittpunkts der
 verschobenen Positionslinie mit der Ekliptik zurueck.
   @param  *m   Pointer auf MUNPOS Instanz
   @param  lon  Laenge
   @param  lat  Breite
   @param  arc  Direktionsbogen
   @return Schiefe Laenge
*/
double mpCalcLongDir(MUNPOS *m, double lon, double lat, double arc);

/**
 === Berechnet die mundane Position aus Laenge und Breite. ===
   @param  *m  Pointer auf MUNPOS Instanz
   @param  lon Laenge
   @param  lat Breite
   @return Mundane Position
*/
double mpCalcPosition(MUNPOS *m, double lon, double lat);

/**
 === Mundane Position ===
 Berechnet die mundane Position aus Rektaszension und Deklination
 wenn flag == 0 oder aus Laenge und Breite, wenn flag == 1.
   @param  \*m   Pointer auf MUNPOS Instanz
   @param  x   Rektaszension oder Laenge
   @param  y   Deklination oder Breite
   @param  flag
   @return Mundane Position
*/
double mpCalcPos2(MUNPOS *m, double x, double y, int flag);

/**
 === Interpolierte mundane Position ===
 Interpoliert die mundane Position linear auf den Ekliptikboegen zwischen 
 den  Haeuserspitzen, wie es in computerlosen Zeiten ueblich war.
 Diese Funktion dient in erster Linie fuer Vergleichszwecke, und sollte
 aufgrund der Ungenauigkeit fuer uebliche Berechnungen nicht verwendet
 werden.
   @param  \*m Pointer auf MUNPOS Instanz
   @param  n   Anzahl der Haeuserspitzen (mindestens 4)
   @param  lon Laenge
   @return Interpolierte mundane Position
*/
double mpCalcPosIP(MUNPOS *m, int n, double lon);

/**
 === Direktion zu mundaner Position ===
 Berechnet den Direktionsbogen aus Laenge und Breite des Promissors,
 mundaner Position des Signifikators, sowie Aspektwinkel/Spiegelpunkt.
 Diese Funktion eignet sich **nicht** fuer die Systeme mpKUEH und mpPOPA,
 da hier die Deklination des Signifikators nicht bekannt ist.
   @param  *m      Pointer auf MUNPOS Instanz
   @param  pro_lon Laenge des Promissors
   @param  pro_lat Breite des Promissors
   @param  mp      Mundane Position des Signifikators
   @param  flags   flags
   @param  asp     Aspektwinkel oder Spiegelpunkt (wenn Flag mpMIRROR gesetzt)
   @flags  mpMUNDO     Direktion zu Mundanaspekt
   @flags  mpMIRROR    Mundanen Spiegelpunkt berechnen
   @flags  mpLAT_P     Promissor mit Breite berechnen
   @flags  mpLAT_S     Signifikator mit Breite berechnen
   @flags  mpBIANC     Promissorbreite nach `Bianchini`, wenn mpLAT_P gesetzt
   @return Direktionsbogen oder im Fehlerfall mpERROR
*/
double mpCalcArc(MUNPOS *m, double pro_lon, double pro_lat, double mp, unsigned flags, double asp);

/**
 === Interplanetare Direktion ===
 Berechnet den Direktionsbogen aus Laenge und Breite des Promissors, Laenge
 und Breite des Signifikators, sowie Aspektwinkel/Spiegelpunkt.
 Mit den Systemen mpKUEH und mpPOPA sind Direktionen
 zu Mundanaspekten **nicht** moeglich.
   @param  *m   Pointer auf MUNPOS Instanz
   @param  pro_lon  Laenge des Promissors
   @param  pro_lat  Breite des Promissors
   @param  sig_lon  Laenge des Signifikators
   @param  sig_lat  Breite des Signifikators
   @param  flags    flags
   @param  asp      Aspektwinkel oder Spiegelpunkt (wenn Flag mpMIRROR gesetzt)
   @flags  mpMUNDO  Direktion zu Mundanaspekt
   @flags  mpMIRROR Mundanen Spiegelpunkt berechnen
   @flags  mpLAT_P  Promissor mit Breite berechnen
   @flags  mpLAT_S  Signifikator mit Breite berechnen
   @flags  mpBIANC  Promissorbreite nach `Bianchini`, wenn mpLAT_P gesetzt
   @return Direktionsbogen oder im Fehlerfall mpERROR
*/
double mpCalcArc2(MUNPOS *m, double pro_lon, double pro_lat, double sig_lon, double sig_lat, unsigned flags, double asp);

/*
 === Rapt Parallele. ===
 Berechnet iterativ den Direktionsbogen aus dem Equidistanzpunkt zweier
 durch Laengen und Breiten gegebenener Promissoren und der mundanen
 Position des Signifikators.
   @param  *m      Pointer auf MUNPOS Instanz
   @param  p1_lon  Laenge Promissor 1
   @param  p1_lat  Breite Promissor 1
   @param  p2_lon  Laenge Promissor 2
   @param  p2_lat  Breite Promissor 2
   @param  mp      Mundane Position des Signifikators
   @param  vect    Pointer auf double[2]
   @retval vect[0] Direktionsbogen
   @retval vect[1] aequidistanz
   @return Anzahl der erforderlichen Iterationsschritte oder im Fehlerfall -1
*/
int mpCalcRapt(MUNPOS *m, double p1_lon, double p1_lat, double p2_lon, double p2_lat, double mp, double *vect);

/**
 === Direktion mundaner aequidistanzen ===
 Berechnet iterativ den Direktionsbogen aus dem aequidistanzpunkt zweier
 durch Laengen und Breiten gegebenener Promissoren und der mundanen
 Position des Signifikators.
   @param  *m      Pointer auf MUNPOS Instanz
   @param  p1_lon  Laenge Promissor 1
   @param  p1_lat  Breite Promissor 1
   @param  p2_lon  Laenge Promissor 2
   @param  p2_lat  Breite Promissor 2
   @param  mp      Mundane Position des Signifikators
   @param  vect    Pointer auf double[6]
   @retval vect[0] Direktionsbogen
   @retval vect[1] aequidistanz
   @retval vect[2] Direktionsbogen zum Oppositionspunkt
   @retval vect[3] aequidistanz
   @retval vect[4] Direktionsbogen zur kleineren aequidistanz
   @retval vect[5] aequidistanz
   @return Anzahl der gesamten Iterationsschritte oder im Fehlerfall -1
*/
int mpCalcArcRapt(MUNPOS *m, double lon1, double lat1, double lon2, double lat2, double mpos, double *vect);

/**
 === Rektaszension aus Laenge und Breite ===
   @param  *m Pointer auf MUNPOS Instanz
   @param lon Laenge
   @param lat Breite
   @return Rektaszension
*/
double mpCalcRA(MUNPOS *m, double lon, double lat);

/**
 === Deklination aus Laenge und Breite ===
   @param  *m Pointer auf MUNPOS Instanz
   @param lon Laenge
   @param lat Breite
   @return Deklination
*/
double mpCalcDE(MUNPOS *m, double lon, double lat);

/**
 === Azimut aus Laenge und Breite ===
   @param  *m Pointer auf MUNPOS Instanz
   @param lon Laenge
   @param lat Breite
   @return Azimut
*/
double mpCalcAZ(MUNPOS *m, double lon, double lat);

/**
 === Hoehe aus Laenge und Breite ===
   @param  *m Pointer auf MUNPOS Instanz
   @param lon Laenge
   @param lat Breite
   @return Hoehe
*/
double mpCalcALT(MUNPOS *m, double lon, double lat);

/*
 === Obere Meridiandistanz ===
   Negativer Stundenwinkel im Gradmass
   @param  *m Pointer auf MUNPOS Instanz
   @param  ra Rektaszension
   @return Obere Meriniandistanz im Bereich +/- 180 Grad
*/
double mpCalcUMD(MUNPOS *m, double ra);

/*
 === Aszensional- / Azimutaldifferenz ===
   @param *m  Pointer auf MUNPOS Instanz
   @param de  Deklination
   @return Aszensional-/Azimutaldifferenz oder im Fehlerfall mpERROR
*/
double mpCalcAD(MUNPOS *m, double de);

/*
 === Halbbogen ===
 Berechnet den Halbbogen, in dem sich das Gestirn befindet.
 Der Nachthalbbogen wird mit negativem Vorzeichen zurueckgegeben.
   @param *m  Pointer auf MUNPOS Instanz
   @param ra  Rektaszension
   @param de  Deklination
   @return Halbbogen oder im Fehlerfall mpERROR
*/
double mpCalcSA(MUNPOS *m, double ra, double de);

/*
 === OA oder OD aus Rektaszension und Deklination ===
 Berechnet abhaengig vom Parameter flag die OA oder OD
   @param *m  Pointer auf MUNPOS Instanz
   @param ra Rektaszension
   @param de Deklination
   @param flag  0 == OA, 1 == OD
   @return OA oder OD, im Fehlerfall mpERROR
*/   
double mpCalcOA(MUNPOS *m, double ra, double de, int flag);

/**
 === OA/OD und Polhoehe aus mundaner Position ===
 Schreibt OA/OD und Polhoehe des der mundanen Position entsprechenden
 Grosskreises in ein Array vom Typ double[2].<br />
 Die Polhoehe kann auch  negative Werte annehmen.
   @param *m     Pointer auf MUNPOS Instanz
   @param mp     mundane Position
   @param \*gc   Pointer auf double[2]
   @retval gc[0] OA
   @retval gc[1] Polhoehe
   @return 0 wenn ok oder im Fehlerfall 1.
*/
int mpCalcGC(MUNPOS *m, double mp, double *gc);

/*
 === OA/OD und Polhoehe aus Rektaszension und Deklination ===
 Schreibt OA/OD und Polhoehe des der mundanen Position entsprechenden
 Grosskreises in ein Array vom Typ double[2].<br />
 Die Polhoehe kann auch  negative Werte annehmen.
   @param *coord Pointer auf aequatoriale Koordinaten
   @param \*gc Pointer auf double[2]
   @retval gc[0] OA
   @retval gc[1] Polhoehe
   @return 0 wenn ok oder im Fehlerfall 1.
*/
double calcGC2(MUNPOS *m, double *coord, double *gc);

/**
 === Berechnet verschiedene Werte fuer das Spekulum ===

 Berechnet verschiedene Werte fuer das Spekulum und schreibt
 sie in eine Instanz der Struktur SPECULUM.
 
 Ist mpPLAC gesetzt, wird die Polhoehe nach `Kuehr` berechnet.
 Die Spekulum Daten unterscheiden sich bei diesen Systemen also nicht.
 Fuer die Direktionssysteme mpKUEH und mpPOPA wird die mundane
 Position nach `Placidus` berechnet.

   @param \*m Pointer auf MUNPOS Instanz
   @param lon ekl. Laenge
   @param lat ekl. Breite
   @param \*spec Pointer auf SPECULUM Instanz

   @retval SPECULUM::lon Kopie der Laenge
   @retval SPECULUM::lat Kopie der Breite
   @retval SPECULUM::ra  Rektaszension
   @retval SPECULUM::de  Deklination
   @retval SPECULUM::az  Azimut (Nullpunkt: Norden)
   @retval SPECULUM::alt Hoehe
   @retval SPECULUM::umd Obere Meridiandistanz im Bereich +/- 180 Grad, im Westen negativ
   @retval SPECULUM::ad  Aszensional- / Azimutaldifferenz
   @retval SPECULUM::sa  Halbbogen, Nachthalbbogen negativ
   @retval SPECULUM::oa  OA oder OD
   @retval SPECULUM::pd  Proportionaldistanz mit Vorzeichen (Meridiandistanz / Halbbogen)
   @retval SPECULUM::mp  Mundane Position
   @retval SPECULUM::q   Aszensionaldifferenz unter eigenem Pol
   @retval SPECULUM::w   OA/OD unter eigenem Pol
   @retval SPECULUM::psi Polhoehe (kann auch negativ sein)
   @retval SPECULUM::md  Meridiandistanz, am IC negativ
   @retval SPECULUM::hd  Horizontdistanz, im Westen negativ
   @retval SPECULUM::zd  Zenit-/Nadirdistanz, am Nadir negativ
   @retval SPECULUM::ed  Ostpunk-/Westpunktdistanz, am Westpunkt negativ
   @retval SPECULUM::rl  Rechte Laenge
*/
void mpCalcSpeculum(MUNPOS *m, double lon, double lat, SPECULUM *spec);


/**-----------------------------------------------------------
// Umrechnungen
//----------------------------------------------------------*/

/**
 === Ekliptik <-> aequator ===
 Rechnet ekliptikale Koordinaten in aequatoriale Koordinaten um
 (und vice versa, wenn eps negativ).
   @param  eps      Schiefe der Ekliptik
   @param  lon      Laenge (Rektaszension)
   @param  lat      Breite (Deklination)
   @param  *coord   Pointer auf double[2]
   @retval coord[0] Rektaszension (Laenge)
   @retval coord[1] Deklination (Breite)
*/
void mpCoTrans(double eps, double x, double y, double *coord);

/**
 === aequator -> Horizont ===
 Rechnet aequatoriale Koordinaten in horizontale Koordinaten um.
   @param  lst    Lokale Sternzeit (RAMC)
   @param  phi    Geographische Breite
   @param  ra     Rektaszension
   @param  de     Deklination
   @param  *coord Pointer auf double[2]
   @retval coord[0] Azimut
   @retval coord[1] Hoehe
*/
void mpEqu2Hor(double lst, double phi, double ra, double de, double *coord);

/**
 === Horizont -> aequator ===
 Rechnet horizontale Koordinaten in aequatoriale Koordinaten um.
   @param  lst    Lokale Sternzeit (RAMC)
   @param  phi    Geographische Breite
   @param  az     Azimut
   @param  alt    Hoehe
   @param  *coord Pointer auf double[2]
   @retval coord[0] Rektaszension
   @retval coord[1] Deklination
*/
void mpHor2Equ(double lst, double phi, double az, double alt, double *equ);

/**
 === In den Bereich 0..360 Grad verschieben ===
   @param alpha  Winkel alpha
   @return 0&deg; &le; alpha &lt; 360&deg;
*/
double mpNorm360(double alpha);

/**
 === In den Bereich -180..+180 Grad verschieben ===
   @param alpha  Winkel alpha
   @return -180&deg; &le; alpha &lt;+180&deg;
*/
double mpNorm180(double alpha);

/**
 === Julianischer Tag JD aus Kalenderdatum ===
 Schreibt die uebergebenen Werte und den daraus errechneten Julianischen Tag
 in eine Instanz der Struktur DATTIM wenn *dt != NULL.
   @param *dt   NULL oder Pointer auf DATTIM Instanz
   @param year  Jahr
   @param month Monat
   @param day   Tag
   @param hours Tagesbruchteil in Stunden
   @param jul   0 == gregorianischer, 1 == julianischer Kalender
   @return Julianischer Tag
*/
double mpCal2JD(int year, int month, int day, double hours, int jul, DATTIM *dt);

/**
 === Kalenderdatum aus julianischem Tag JD ===
 Schreibt den uebergebenen Wert sowie das daraus errechnete
 Kalenderdatum in eine Instanz der Struktur DATTIM.
   @param *dt Pointer auf DATTIM Instanz
   @param jd  Julianischer Tag
   @retval year  Jahr
   @retval month Monat
   @retval day   Tag
   @retval hours Tagesbruchteil in Stunden
*/
void mpJD2Cal(double jd, DATTIM *dt);


/**-----------------------------------------------------------
// Verschiedenes
//----------------------------------------------------------*/

/*
 === Differenz aus a, b (Im Uhrzeigersinn negativ) ===
*/
double mpDif(double a, double b);

/*
 === Differenz a, b (Immer positiv) ===
*/
double mpADif(double a, double b);

/*
 === Arithmetischer Mittelpunkt ===
*/
double mpMid(double a, double b);

/**
 === Geometrischer Aspektwinkel (Orthodrome) und Halbdistanz ===
 Berechnet den Bogen der Orthodrome. 
 
 Wenn *vect != NULL werden Laenge und Breite des geometrischen 
 Mittelpunkts berechnet und in das Array vect geschrieben.

   @param L1 Laenge Startpunkt
   @param B1 Breite Startpunkt
   @param L2 Laenge Endpunkt
   @param B2 Breite Endpunkt
   @param *vect NULL oder Pointer auf double[2]
   @retval vect[0] Laenge (optional)
   @retval vect[1] Breite (optional)
   @return Bogen der Orthodrome
*/
double mpAspect(double L1, double B1, double L2, double B2, double *vect);

/**
 === Geozentrische Breite aus geographischer Breite ===
   @param phi  Geographische Breite
   @return Geozentrische Breite
*/
double mpGeoZ(double phi);

/*
 === Aszensionaldifferenz aus Polhoehe und Deklination ===
   @param psi Geographische Breite
   @param de  Polhoehe
   @param de  Deklination
   @return Aszensionaldifferenz oder im Fehlerfall mpERROR
*/
double mpAD(double psi, double de);

/*
 === OA/OD aus Rektaszension, Deklination und Polhoehe ===
 Berechnet abhaengig vom Parameter flag die OA oder OD
   @param *m    Pointer auf MUNPOS Instanz
   @param ra    Rektaszension
   @param de    Deklination
   @param psi   Polhoehe
   @param flag  0 == OA, 1 == OD
   @return OA oder OD, im Fehlerfall mpERROR
*/   
double mpOA(MUNPOS *m, double ra, double de, double psi, int flag);

/*
 === Laenge aus Schiefe der Ekliptic, OA und Polhoehe ===
   @param eps Schiefe der Ekliptik
   @param oa  Schiefe Aufsteigung
   @param psi Polhoehe
   @return Laenge
*/
double mpLongGC(double eps, double oa, double psi);

#ifdef __cplusplus
}
#endif

#endif
